classdef VEMalgo < handle
    %==========================================================================
    % VEM algorithm for multichannel audio source separation with MDCT
    % source representation
    %==========================================================================

    properties
        x; % time-domain observations [T x I]

        % Model parameters
        a; % mixing filters [I x J x La]
        sigmaNoise; % additive noise variances vector [1 x I]
        W; % {J x 1} cell array containing the matrices of NMF bases for each source
        H; % {J x 1} cell array containing the matrices of NMF activations for each source

        % Variational distribution parameters
        m_post; % mean [F x N x J]
        gamma_post; % variance [F x N x J]

        % Others
        expectError; % expectation of the squared error between the model and the observations (sum_t e_i(t)) [1 x I]
        varFreeEnergy; % objective
        psiCorr; % troncated autocorrelation of time-frequency MDCT atoms
        G; % [F x (La + wlen - 1) x I x J] : contains the convolution of the MDCT atom at frequency f with the mixing filters a_ij
        s; % [Ls x J] time-domain estimated sources
        sim; % [T x J x I] time-domain estimated source images
    end

    methods

        function this = VEMalgo(varargin)
            this.Set(varargin{:});
        end

        function Set(this,varargin)
            st_input = struct(varargin{:});

            if isfield(st_input, 'x')
                this.x = st_input.x;
            end

            if isfield(st_input, 'a')
                this.a = st_input.a;
            end

            if isfield(st_input, 'sigmaNoise')
                this.sigmaNoise = st_input.sigmaNoise;
            end

            if isfield(st_input, 'W')
                this.W = st_input.W;
            end

            if isfield(st_input, 'H')
                this.H = st_input.H;
            end

            if isfield(st_input, 'psi')
                this.H = st_input.psi;
            end

            J = numel(this.W);
            F = size(this.W{1},1);
            N = size(this.H{1},2);
            wlen = 2*F;
            this.m_post = zeros(F,N,J);
            this.gamma_post = zeros(F,N,J);

            [T, I] = size(this.x);
            La = size(this.a,3);
            Ls = T - La + 1;
            this.s = zeros(Ls, J);
            this.sim = zeros(T, J, I);
            this.ComputePsiCorr;
            this.G = zeros(F, La + wlen - 1, I, J);
            this.ComputeGMat;
            this.expectError = zeros(1,I);
        end

        function [] = ComputePsiCorr(this)

            F = size(this.W{1},1);
            [T, ~] = size(this.x);
            La = size(this.a,3);
            Ls = T - La + 1;
            wlen = 2*F;
            psiCorrTmp = zeros(F,max(wlen,La));
            for f=1:F
                psi_fn = this.GetMDCTAtom(f-1,0,Ls,wlen);
                tmp = xcorr(psi_fn(1:max(wlen,La)));
                psiCorrTmp(f,:) = tmp(max(wlen,La):end).';
            end
            this.psiCorr = psiCorrTmp;
        end

        function [] = EStepPCG(this)
            fprintf('E-step\n');

            [~, I] = size(this.x);
            J = numel(this.W);
            F = size(this.W{1},1);
            N = size(this.H{1},2);
            La = size(this.a,3);
            wlen = 2*F;
            hop = wlen/2;

            v_prior = zeros(F,N,J);

            % compute a priori source variances
            for j = 1:J
                v_prior(:,:,j) = this.W{j}*this.H{j};
            end

            % compute gamma_post
            tmp = squeeze(sum(sum(this.G.^2,2)./permute(repmat(this.sigmaNoise', [1, 1, J, F]), [4,2,1,3]),3)); % F x J
            tmp = repmat(tmp, [1,1,N]);
            tmp = permute(tmp, [1,3,2]);
            this.gamma_post = 1./(1./v_prior + tmp);

            invD = diag(sparse(this.gamma_post(:)));

            % compute gradient
            epsilon = this.x - sum(permute(this.sim,[1,3,2]),3); % T x I
            frameInd = zeros(wlen+La-1,N); % column n of frameInd contains the indices of the signal to be considered for frame n
            for n=1:N
                frameInd(:,n) = (n-1)*hop + (1:wlen+La-1)';
            end
            epsilonMat = zeros(wlen+La-1,N,J,I);
            for i=1:I
                epsilon_i = epsilon(:,i)/this.sigmaNoise(i); % (wlen + La - 1) x N
                epsilonMat(:,:,:,i) = repmat(epsilon_i(frameInd), [1,1,J]); % (wlen + La - 1) x N x J x I
            end

            this.G = permute(this.G, [1,2,4,3]);
            grad = this.m_post./(v_prior+eps) - sum(mtimesx(this.G,epsilonMat),4);
            this.G = permute(this.G, [1,2,4,3]);

            % compute descent direction
            w = invD*grad(:);
            w = reshape(w, F, N, J);

            niter = 10;
            for iter = 1:niter

                % compute gamma
                w_mdct = this.ImdctMulti(w); % Ls x J
                w_filt = this.SrcImage(w_mdct); % T x J x I
                w_filt = squeeze(sum(w_filt, 2)); % T x I
                w_filtMat = zeros(wlen+La-1,N,J,I); % (wlen + La - 1) x N x J x I
                for i=1:I
                    w_filt_i = w_filt(:,i)/this.sigmaNoise(i); % (wlen + La - 1) x N
                    w_filtMat(:,:,:,i) = repmat(w_filt_i(frameInd), [1,1,J]);
                end

                this.G = permute(this.G, [1,2,4,3]);
                gamma = w./(v_prior+eps) + sum(mtimesx(this.G,w_filtMat),4);
                this.G = permute(this.G, [1,2,4,3]);

                % compute mu
                mu = real((w(:)'*grad(:))/(w(:)'*gamma(:)));

                % update parameters to be estimated
                this.m_post = this.m_post - mu*w;

                % update source images
                this.ImdctMultiProp(1:J);
                this.SrcImageProp(1:J,(1:I));

                if iter == niter
                    break;
                end

                % compute gradient
                epsilon = this.x - sum(permute(this.sim,[1,3,2]),3); % T x I
                epsilonMat = zeros(wlen+La-1,N,J,I);
                for i=1:I
                    epsilon_i = epsilon(:,i)/this.sigmaNoise(i); % (wlen + La - 1) x N
                    epsilonMat(:,:,:,i) = repmat(epsilon_i(frameInd), [1,1,J]); % (wlen + La - 1) x N x J
                end

                this.G = permute(this.G, [1,2,4,3]);
                grad = this.m_post./(v_prior+eps) - sum(mtimesx(this.G,epsilonMat),4);
                this.G = permute(this.G, [1,2,4,3]);

                % preconditionning
                gradP = invD*grad(:);
                gradP = reshape(gradP, F, N, J);

                % compute alpha
                alpha = -real((gamma(:)'*gradP(:))/(w(:)'*gamma(:)));

                % compute w
                w = gradP + alpha*w;

            end

        end

        function [] = UpdateNMF(this)
            J = numel(this.W);
            for j=1:J
                [this.W{j}, this.H{j}, ~] = this.ISNMF(this.m_post(:,:,j).^2 + this.gamma_post(:,:,j), this.W{j}, this.H{j}, 50);
            end
        end

        function [] = UpdateNoiseVar(this)
            [T,~] = size(this.x);
            this.sigmaNoise = 1/T*this.expectError;
        end

        function [] = ComputeExpectError(this)

            epsilon = this.x - squeeze(sum(this.sim,2)); % T x I
            
            normG = permute(sum(this.G.^2,2), [1,2,4,3]); % F x 1 x J x I
            gamma_normG = bsxfun(@times, this.gamma_post, normG); % F x N x J x I
            
            this.expectError = sum(epsilon.^2,1) + permute(sum(sum(sum(gamma_normG,1),2),3), [3,4,1,2]);
        end

        function [W, H, E] = ISNMF(this, V, W0, H0, iter_num)

            if nargin < 4 || isempty(iter_num)
                iter_num = 100;
            end

            [F,N] = size(V);

            W = W0;
            H = H0;

            S = sum(V(:));
            matOfOnes = ones(F,N);
            WH = W*H + eps;

            E = zeros(1, iter_num);

            for iter = 1:iter_num
                % update of W
                W = W .* (( V./((WH).^2) )*H' + eps) ./ ((1./(WH))*H' + eps);

                % normalization
                w = sum(W,1);
                d = diag(1./w);
                W = W*d; % Normalizatiosn of columns of W
                d = diag(w);
                H = d*H; % Energy transfer to H

                WH = W*H + eps;

                % update of H
                H = H .* (W'*(V./((WH).^2)) + eps) ./ (W'*(1./(WH)) + eps);
                WH = W*H + eps;

                % Error with IS divergence:
                E(iter) = sum(sum(V./(WH + eps) - log(V./(WH + eps) + eps) - matOfOnes))/S; % normalized error
            end
        end

        function [] = ComputeVarFreeEnergy(this, b_computeExpectError)

            if nargin < 2
                b_computeExpectError = 1;
            end

            [T,~] = size(this.x);
            J = numel(this.W);

            if b_computeExpectError
                this.ComputeExpectError;
            end

            cost = -0.5*sum(this.expectError./this.sigmaNoise + T*log(this.sigmaNoise));

            for j=1:J
                Vj = this.W{j}*this.H{j};
                cost = cost - 0.5*sum(sum( log(Vj./(this.gamma_post(:,:,j)+eps) + eps) + (this.m_post(:,:,j).^2 + this.gamma_post(:,:,j))./(Vj + eps) ));
            end

            this.varFreeEnergy = cost;

        end

        function [ psi ] = GetMDCTAtom(this, f, n, siglen, wlen )

            F = wlen/2;
            win = sin((.5:wlen-.5)/wlen*pi); % sine window
            hop = wlen/2; % hop size = wlen/2;

            frameInd = (n*hop+1 : n*hop + wlen)';
            if frameInd(end)>siglen
                error('out of signal')
            end
            psi = zeros(siglen,1);
            psi(frameInd) = win.*( sqrt(2/F) * cos( (f+1/2) * ((0:wlen-1) + 0.5 + wlen/4) * 2*pi/wlen ) );

        end

        function [] = ComputeGMat(this)

            F = size(this.m_post,1);
            [~,I] = size(this.x);
            J = numel(this.W);
            wlen = 2*F;
            win = sin((.5:wlen-.5)/wlen*pi);

            for f=0:F-1
                for j=1:J
                    psi_f = win.*( sqrt(2/F) * cos( (f+1/2) * ((0:wlen-1) + 0.5 + wlen/4) * 2*pi/wlen ) ); % wlen x 1
                    a_sq = permute(this.a(:,j,:),[3,1,2]); % La x I
                    this.G(f+1,:,:,j) = this.FFTConv(psi_f'*ones(1,I),a_sq); % (La + wlen - 1) x I
                end
            end
        end

        function [] = ImdctMultiProp(this, srcInd)

            [F, N, J] = size(this.m_post);
            Ls = size(this.s,1);

            if nargin<2
                srcInd = 1:J;
            end

            wlen = 2*F;
            hop = wlen/2;
            win = sin((.5:wlen-.5)/wlen*pi);

            X_ext = [this.m_post(:,:,:) ; -this.m_post(end:-1:1,:,:)];
            x_rec = zeros(hop*(N+1), 1);

            for j = srcInd
                x_rec(:) = 0;
                frames = real(ifft(X_ext(:,:,j).*(exp(1i*2*pi*(0.5+wlen/4)*(0:wlen-1)'/wlen)*ones(1,N)))...
                    .*(exp(1i*pi*(0.5+wlen/4+(0:wlen-1)')/wlen)*ones(1,N)))*sqrt(wlen);
                for n = 1:N
                    frameInd = (n-1)*hop + (1:wlen);
                    x_rec(frameInd) = x_rec(frameInd) + win'.*frames(:,n);
                end
                this.s(:,j) = x_rec(1:Ls,:,:);
            end
        end

        function [x] = ImdctMulti(this, X)

            [F, N, J] = size(X);
            Ls = size(this.s,1);
            x = zeros(Ls,J);

            wlen = 2*F;
            hop = wlen/2;
            win = sin((.5:wlen-.5)/wlen*pi);

            X_ext = [X(:,:,:) ; -X(end:-1:1,:,:)];
            x_rec = zeros(hop*(N+1), 1);

            for j = 1:J
                x_rec(:) = 0;
                frames = real(ifft(X_ext(:,:,j).*(exp(1i*2*pi*(0.5+wlen/4)*(0:wlen-1)'/wlen)*ones(1,N)))...
                    .*(exp(1i*pi*(0.5+wlen/4+(0:wlen-1)')/wlen)*ones(1,N)))*sqrt(wlen);
                for n = 1:N
                    frameInd = (n-1)*hop + (1:wlen);
                    x_rec(frameInd) = x_rec(frameInd) + win'.*frames(:,n);
                end
                x(:,j) = x_rec(1:Ls,:,:);
            end
        end

        function [] = SrcImageProp(this, srcInd, micInd)
            % s : [Ls x J]
            % si : [T x J x I]

            J = length(srcInd);
            I = length(micInd);
            T = size(this.x,1);
            Ls = size(this.s,1);
            La = size(this.a,3);

            s = this.s(:, srcInd);
            a = this.a(micInd, srcInd, :);

            s_resh = zeros(Ls,J,I);
            for i=1:I
                s_resh(:,:,i) = s;
            end
            s_resh = reshape(s_resh, Ls, I*J);
            a_resh = reshape(permute(a,[3,2,1]), La, I*J);
            si = this.FFTConv(s_resh,a_resh);
            this.sim(:,srcInd, micInd) = reshape(si, T, J, I);

        end

        function [xi] = SrcImage(this, x)
            % s : [Ls x J]
            % si : [T x J x I]
            J = size(x,2);
            [T, I] = size(this.x);
            Ls = size(this.s,1);
            La = size(this.a,3);

            x_resh = zeros(Ls,J,I);
            for i=1:I
                x_resh(:,:,i) = x;
            end
            x_resh = reshape(x_resh, Ls, I*J);
            a_resh = reshape(permute(this.a,[3,2,1]), La, I*J);
            xi = this.FFTConv(x_resh,a_resh);
            xi = reshape(xi, T, J, I);

        end

        function y = FFTConv(this, x, h)
            T = length(x) + length(h) - 1;
            Nfft = 2^nextpow2(T);
            y = real(ifft(fft(x, Nfft).*fft(h, Nfft), Nfft));
            y = y(1:T,:);
        end

    end
end
