classdef VEMalgo < handle
    %==========================================================================
    % VEM algorithm for multichannel audio source separation with OFSTFT
    % source representation
    %==========================================================================
    
    properties
        % Model parameters
        x; % time domaine observations [T x I]
        a; % mixing filters [I x J x La]
        sigmaNoise; % additive noise variances vector [1 x I]
        W; % {J x 1} cell array containing the matrices of NMF bases for each source
        H; % {J x 1} cell array containing the matrices of NMF activations for each source
        
        % Variational distribution parameters
        m_post_r; % mean of real part [F x N x J]
        m_post_i; % mean of imaginary part [F x N x J]
        gamma_post_r; % variance of real part [F x N x J]
        gamma_post_i; % variance of imaginary part [F x N x J]
        rho_post; % correlation coefficient between real and imaginary parts [F x N x J]
        
        % Others
        expectError; % expectation of the squared error between the model and the observations (sum_t e_i(t)) [1 x I]
        varFreeEnergy; % objective
        psiCorr; % troncated autocorrelation of time-frequency OFSTFT atoms
        G; % [F x (La + wlen - 1) x I x J] : contains the convolution of the OFSTFT atom at frequency f with the mixing filters a_ij
        s; % [Ls x J] time-domain sources
        sim; % [T x J x I] time-domain source images
        win; % [1 x wlen] OFSTFT window
        hop_percent; % hop size as a percentage of the window length
    end
    
    methods
        
        function this = VEMalgo(varargin)
            this.Set(varargin{:});
        end
        
        function Set(this,varargin)
            st_input = struct(varargin{:});
            
            if isfield(st_input, 'x')
                this.x = st_input.x;
            end
            
            if isfield(st_input, 'a')
                this.a = st_input.a;
            end
            
            if isfield(st_input, 'sigmaNoise')
                this.sigmaNoise = st_input.sigmaNoise;
            end
            
            if isfield(st_input, 'W')
                this.W = st_input.W;
            end
            
            if isfield(st_input, 'H')
                this.H = st_input.H;
            end
            
            if isfield(st_input, 'psi')
                this.H = st_input.psi;
            end
            
            J = numel(this.W);
            F = size(this.W{1},1);
            N = size(this.H{1},2);
            wlen = 2*F;
            
            if isfield(st_input, 'win')
                this.win = st_input.win;
            else
                this.win = sin((.5:wlen-.5)/wlen*pi);
            end
            
            if isfield(st_input, 'hop_percent')
                this.hop_percent = st_input.hop_percent;
            else
                this.hop_percent = 0.5;
            end
                        
            this.m_post_r = zeros(F,N,J);
            this.m_post_i = zeros(F,N,J);
            this.gamma_post_r = zeros(F,N,J);
            this.gamma_post_i = zeros(F,N,J);
            this.rho_post = zeros(F,N,J);
            
            [T, I] = size(this.x);
            La = size(this.a,3);
            Ls = T - La + 1;
            this.s = zeros(Ls, J);
            this.sim = zeros(T, J, I);
            this.ComputePsiCorr;
            this.G = zeros(F, La + wlen - 1, I, J);
            this.ComputeGMat;
            this.expectError = zeros(1,I);
        end
        
        function [] = ComputePsiCorr(this)
            
            F = size(this.W{1},1);
            [T, ~] = size(this.x);
            La = size(this.a,3);
            Ls = T - La + 1;
            wlen = 2*F;            
            psiCorrTmp = zeros(F,max(wlen,La));
            for f=1:F
                psi_fn = this.GetGSTFTAtom(f-1,0,Ls);
                tmp = xcorr(psi_fn(1:max(wlen,La)));
                psiCorrTmp(f,:) = tmp(max(wlen,La):end).'; 

            end
            this.psiCorr = psiCorrTmp;
            
        end
               
        function [] = EStepPCGComplex(this)
            fprintf('E-step\n');
                        
            [~, I] = size(this.x);
            J = numel(this.W);
            F = size(this.W{1},1);
            N = size(this.H{1},2);
            La = size(this.a,3);
            wlen = 2*F;
            hop = wlen*this.hop_percent;
            
            v_prior = zeros(F,N,J);
       
            % compute a priori source variances
            for j = 1:J
                v_prior(:,:,j) = this.W{j}*this.H{j};
            end
            
            % compute rho_post
            den_r = squeeze( 2 * sum(sum(real(this.G).^2,2)...
                ./permute(repmat(this.sigmaNoise', [1, 1, J, F]), [4,2,1,3]),3)); % F x J
            den_r = repmat(den_r, [1,1,N]);
            den_r = permute(den_r, [1,3,2]);
            den_r = 1./(v_prior) + den_r;
            
            den_i = squeeze( 2 * sum(sum(imag(this.G).^2,2)...
                ./permute(repmat(this.sigmaNoise', [1, 1, J, F]), [4,2,1,3]),3)); % F x J
            den_i = repmat(den_i, [1,1,N]);
            den_i = permute(den_i, [1,3,2]);
            den_i = 1./(v_prior) + den_i;
            
            num = squeeze( 2 * sum(sum(real(this.G).*imag(this.G),2)...
                ./permute(repmat(this.sigmaNoise', [1, 1, J, F]), [4,2,1,3]),3)); % F x J
            num = repmat(num, [1,1,N]);
            num = permute(num, [1,3,2]);
            
            this.rho_post = num./(den_r.^(.5) .* den_i.^(.5));
            
            clear num;
            
            % compute gamma_post
            this.gamma_post_r = 1./( 2*(1 - this.rho_post.^2) .* den_r );
            this.gamma_post_i = 1./( 2*(1 - this.rho_post.^2) .* den_i );
            
            % compute inverse of the preconditionning matrix
            invD = diag( sparse( (1 - this.rho_post(:).^2) ./ (1./this.gamma_post_r(:) + 1./this.gamma_post_i(:)) ) );
            
            clear den_r den_i;

            % compute gradient
            epsilon = this.x - sum(permute(this.sim,[1,3,2]),3); % T x I
            frameInd = zeros(wlen+La-1,N); % column n of frameInd contains the indices of the signal to be considered for frame n
            for n=1:N
                frameInd(:,n) = (n-1)*hop + (1:wlen+La-1)';
            end
            epsilonMat = zeros(wlen+La-1,N,J,I);
            for i=1:I
                epsilon_i = epsilon(:,i)/this.sigmaNoise(i); % (wlen + La - 1) x N
                epsilonMat(:,:,:,i) = repmat(epsilon_i(frameInd), [1,1,J]); % (wlen + La - 1) x N x J x I
            end
            
            this.G = permute(this.G, [1,2,4,3]);
            grad = 2 * ( (this.m_post_r + 1i*this.m_post_i)./(v_prior+eps) - sum(mtimesx(conj(this.G),epsilonMat),4) );
            this.G = permute(this.G, [1,2,4,3]);
            
            % compute descent direction
            w = invD*grad(:);
            w = reshape(w, F, N, J);
            
            niter = 10;
            for iter = 1:niter
                % compute gamma
                w_gstft = this.IofstftMulti(w); % Ls x J
                w_filt = this.SrcImage(w_gstft); % T x J x I
                w_filt = squeeze(sum(w_filt, 2)); % T x I
                w_filtMat = zeros(wlen+La-1,N,J,I); % (wlen + La - 1) x N x J x I
                for i=1:I
                    w_filt_i = w_filt(:,i)/this.sigmaNoise(i); % (wlen + La - 1) x N
                    w_filtMat(:,:,:,i) = repmat(w_filt_i(frameInd), [1,1,J]); 
                end
                
                this.G = permute(this.G, [1,2,4,3]);
                gamma = 2 * ( w./(v_prior+eps) + sum(mtimesx(conj(this.G),w_filtMat),4) );
                this.G = permute(this.G, [1,2,4,3]);
                
                % compute mu
                mu = real((w(:)'*grad(:))/(w(:)'*gamma(:)));
                if mu<0
                	mu = eps;
                end
                
                % update parameters to be estimated
                this.m_post_r = this.m_post_r - mu*real(w);
                this.m_post_i = this.m_post_i - mu*imag(w);

                % update source images
                this.IofstftMultiProp(1:J);
                this.SrcImageProp(1:J,(1:I));
                
                if iter == niter
                    break;
                end

                % compute gradient
                epsilon = this.x - sum(permute(this.sim,[1,3,2]),3); % T x I
                frameInd = zeros(wlen+La-1,N); % column n of frameInd contains the indices of the signal to be considered for frame n
                for n=1:N
                    frameInd(:,n) = (n-1)*hop + (1:wlen+La-1)';
                end
                epsilonMat = zeros(wlen+La-1,N,J,I);
                for i=1:I
                    epsilon_i = epsilon(:,i)/this.sigmaNoise(i); % (wlen + La - 1) x N
                    epsilonMat(:,:,:,i) = repmat(epsilon_i(frameInd), [1,1,J]); % (wlen + La - 1) x N x J x I
                end
                
                this.G = permute(this.G, [1,2,4,3]);
                grad = 2 * ( (this.m_post_r + 1i*this.m_post_i)./(v_prior+eps) - sum(mtimesx(conj(this.G),epsilonMat),4) );
                this.G = permute(this.G, [1,2,4,3]);
                
                % preconditionning
                gradP = invD*grad(:);
                gradP = reshape(gradP, F, N, J);

                % compute beta
                beta = -real((gamma(:)'*gradP(:))/(w(:)'*gamma(:)));
                if beta<0
                    beta = eps;
                end
                
                % compute w
                w = gradP + beta*w;
                
            end
            
        end 
                             
        function [] = UpdateNMF(this)
            J = numel(this.W);
            for j=1:J
                [this.W{j}, this.H{j}, ~] = this.ISNMF(this.m_post_r(:,:,j).^2 + ...
                    this.m_post_i(:,:,j).^2 + this.gamma_post_r(:,:,j) + this.gamma_post_i(:,:,j)  + eps, ...
                    this.W{j}, this.H{j}, 50);
            end
        end
        
        function [] = UpdateNoiseVar(this)
            [T,~] = size(this.x);
            this.sigmaNoise = 1/T*this.expectError;
        end
                
        function [] = ComputeExpectError(this)            
                        
            epsilon = this.x - squeeze(sum(this.sim,2)); % T x I

            normG = permute(sum(abs(this.G).^2,2), [1,2,4,3]); % F x 1 x J x I
            gamma_post = this.gamma_post_r + this.gamma_post_i; % F x N x J
            gamma_normG = bsxfun(@times, gamma_post, normG); % F x N x J x I
            
            G2 = permute(sum(this.G.^2,2), [1,2,4,3]); % F x 1 x J x I
            gamma_post_bar = this.gamma_post_r - this.gamma_post_i + ...
                2*1i*this.rho_post.*sqrt(this.gamma_post_r.*this.gamma_post_i); % F x N x J
            gamma_G2 = bsxfun(@times, gamma_post_bar, G2); % F x N x J x I

            this.expectError = sum(epsilon.^2,1) + 2*permute(sum(sum(sum(gamma_normG + real(gamma_G2),1),2),3),...
                [3,4,1,2]);
        end
        
        function [W, H, E] = ISNMF(this, V, W0, H0, iter_num)
            
            if nargin < 4 || isempty(iter_num)
                iter_num = 100;
            end
            
            [F,N] = size(V);
            
            W = W0;
            H = H0;
            
            S = sum(V(:));
            matOfOnes = ones(F,N);
            WH = W*H + eps;
            
            E = zeros(1, iter_num);
            
            for iter = 1:iter_num
                % update of W
                W = W .* (( V./((WH).^2) )*H' + eps) ./ ((1./(WH))*H' + eps);
                
                % normalization
                w = sum(W,1);
                d = diag(1./w);
                W = W*d; % Normalizatiosn of columns of W
                d = diag(w);
                H = d*H; % Energy transfer to H
                
                WH = W*H + eps;
                
                % update of H
                H = H .* (W'*(V./((WH).^2)) + eps) ./ (W'*(1./(WH)) + eps);
                WH = W*H + eps;
                
                % Error with IS divergence:
                E(iter) = sum(sum(V./(WH + eps) - log(V./(WH + eps) + eps) - matOfOnes))/S; % normalized error
                
            end
        end
        
        function [] = ComputeVarFreeEnergy(this, b_computeExpectError)            
            
            if nargin < 2
                b_computeExpectError = 1;
            end
            
            [T,~] = size(this.x);
            J = numel(this.W);
            
            if b_computeExpectError
                this.ComputeExpectError;
            end
            
            cost = -0.5*sum(this.expectError./this.sigmaNoise + T*log(this.sigmaNoise));
            
            for j=1:J
                Vj = this.W{j}*this.H{j};
                cost = cost - sum(sum( log(Vj + eps) + (this.m_post_r(:,:,j).^2 + this.m_post_i(:,:,j).^2 ...
                    + this.gamma_post_r(:,:,j) + this.gamma_post_i(:,:,j))./(Vj + eps) ));
                
                cost = cost + 0.5*sum(sum( log(this.gamma_post_r(:,:,j) + eps) + log(this.gamma_post_i(:,:,j) + eps) ...
                    + log(1 - this.rho_post(:,:,j).^2 + eps) ));
            end
            
            this.varFreeEnergy = cost;
            
        end
        
        function [ psi ] = GetGSTFTAtom(this, f, n, siglen )

            wlen = length(this.win);
            hop = wlen*this.hop_percent;
            
            frameInd = (n*hop+1 : n*hop + wlen)';
            if frameInd(end)>siglen
                error('out of signal')
            end
            psi = zeros(siglen,1);
            psi(frameInd) = this.win.*( sqrt(1/wlen) * exp(1i*2*pi*(0:wlen-1)*(f+0.5)/wlen) );
            
        end

        function [] = ComputeGMat(this)
            
            F = size(this.m_post_r,1);
            [~,I] = size(this.x);
            J = numel(this.W);
            wlen = 2*F; % window length
            
            for f=0:F-1
                for j=1:J
                    psi_f = this.win.*( sqrt(1/wlen) * exp(-1i*2*pi*(0:wlen-1)*(f+0.5)/wlen) ); % wlen x 1
                    a_sq = permute(this.a(:,j,:),[3,1,2]); % La x I
                    this.G(f+1,:,:,j) = this.FFTConv(psi_f'*ones(1,I),a_sq); % (La + wlen - 1) x I
                end
            end
        end        
        
        function [] = IofstftMultiProp(this, srcInd)

            [F, N, J] = size(this.m_post_r);
            
            if nargin<2
                srcInd = 1:J;
            end
            
            wlen = 2*F; % window length
            hop = wlen*this.hop_percent; % hop size
            
            x_rec = zeros(hop*(N-1) + wlen, 1);
            
            for j = srcInd
                x_rec(:) = 0;
                y = sqrt(wlen)*real( ifft( cat(1, this.m_post_r(:,:,j) + 1i*this.m_post_i(:,:,j),...
                    this.m_post_r(end:-1:1,:,j) - 1i*this.m_post_i(end:-1:1,:,j) ) )...
                    .*(exp(1i*pi*(0:wlen-1)/wlen).' * ones(1,N)));
                for n = 1:N
                    frameInd = (n-1)*hop + (1:wlen);
                    x_rec(frameInd) = x_rec(frameInd) + this.win'.*y(:,n);
                end
                this.s(:,j) = x_rec;
            end
        end
        
        function [x] = IofstftMulti(this, X)

            [F, N, J] = size(X);
            Ls = size(this.s,1);
            x = zeros(Ls,J);
            
            wlen = 2*F; % window length
            hop = wlen*this.hop_percent; % hop size
                      
            X = cat(1,X,conj(X(end:-1:1,:,:)));
                        
            for j = 1:J
                y = sqrt(wlen)*real(ifft(X(:,:,j)).*(exp(1i*pi*(0:wlen-1)/wlen).'*ones(1,N)));
                for n = 1:N
                    frameInd = (n-1)*hop + (1:wlen);
                    x(frameInd,j) = x(frameInd,j) + this.win'.*y(:,n);
                end
            end
        end

        function [] = SrcImageProp(this, srcInd, micInd)
            % s : [Ls x J]
            % si : [T x J x I]
            
            J = length(srcInd);
            I = length(micInd);
            T = size(this.x,1);
            Ls = size(this.s,1);
            La = size(this.a,3);
            
            s = this.s(:, srcInd);
            a = this.a(micInd, srcInd, :);
            
            s_resh = zeros(Ls,J,I);
            for i=1:I
                s_resh(:,:,i) = s;
            end
            s_resh = reshape(s_resh, Ls, I*J);
            a_resh = reshape(permute(a,[3,2,1]), La, I*J);
            si = real(this.FFTConv(s_resh,a_resh));
            this.sim(:,srcInd, micInd) = reshape(si, T, J, I);
            
        end
        
        function [xi] = SrcImage(this, x)
            % s : [Ls x J]
            % si : [T x J x I]
            J = size(x,2);
            [T, I] = size(this.x);
            Ls = size(this.s,1);
            La = size(this.a,3);
                                    
            x_resh = zeros(Ls,J,I);
            for i=1:I
                x_resh(:,:,i) = x;
            end
            x_resh = reshape(x_resh, Ls, I*J);
            a_resh = reshape(permute(this.a,[3,2,1]), La, I*J);
            xi = real(this.FFTConv(x_resh,a_resh));
            xi = reshape(xi, T, J, I);
            
        end
  
        function y = FFTConv(this, x, h)
            T = length(x) + length(h) - 1;  
            Nfft = 2^nextpow2(T);   
            y = ifft(fft(x, Nfft).*fft(h, Nfft), Nfft);
            y = y(1:T,:);
        end
        
    end
end

