%==========================================================================
% Main script
%==========================================================================

addpath('../aux_tools')
addpath('../mtimesx')

%% Parameters

dataDir = '../sounds'; % directory of true source images and mix 
resultsDir = dataDir; % directory to write estimated source images
resultsPrefix = 'src_VEM_OFSTFT_';

J = 5; % number of sources
I = 2; % number of microphones
wlen = 2048; % window length
F = wlen/2; % number of frequency bins
hop_percent = 0.5; % hop size in percentage of the window length
hop = wlen*hop_percent; % hop size in sample
win = createWindow( wlen, hop_percent); % create TF analysis/synthesis window
niter = 100; % number of iterations for the VEM algorithm

wavFile = fullfile(dataDir,'mix.wav'); 
[x, fs] = audioread(wavFile);
X = ofstft_multi(x, win, hop_percent);

T_orig = size(x,1);
T60 = 470; % reverberation time in ms
La = T60*1e-3*fs;
Ls = T_orig-La+1;

%% Zero padding for handling edges

Nzp_beg = wlen*(1-hop_percent);
x = [zeros(Nzp_beg,I) ; x];
Ls = Ls + Nzp_beg;

N = fix(Ls/hop) + (mod(Ls, hop) ~= 0); % number of time frames

Ls_zp = (N-1)*hop + wlen;
Nzp_end = Ls_zp - Ls;
x = [x ; zeros(Nzp_end, I)]; % zero-padding for last frame
T = size(x,1);

%% Initialization

st = load(fullfile(dataDir, 'mix_filters.mat'));
a_init = st.a;

mix_psd = 0.5 * (mean(abs(X(:,:,1)).^2 + abs(X(:,:,2)).^2, 2));
W_init = cell(J,1);
H_init = cell(J,1);
K = 10; % NMF order
for j=1:J
    rng(1);
    W_init{j} = 0.5 * (abs(randn(F,K)) + ones(F,K)) .* (mix_psd * ones(1,K));
    rng(1);
    H_init{j} = 0.5 * (abs(randn(K,N)) + ones(K,N));
end

sigmaNoise_init = sum(abs(x).^2,1)/T;

%% Run VEM algorithm

fprintf('Instantiation of VEMalgo\n');
ss_algoVEM = VEMalgo('x', x, 'a', a_init, 'sigmaNoise', sigmaNoise_init, 'W', {W_init}, 'H', {H_init}, ...
        'win', win, 'hop_percent', hop_percent);
    
varFreeEnergy = zeros(niter,1);

for iter = 1:niter
    fprintf('\nVEM iteration %d of %d\n\n', iter, niter);
    
    ss_algoVEM.EStepPCGComplex;
        
    fprintf('Update NMF\n');
    ss_algoVEM.UpdateNMF;
    
    fprintf('Compute expected error\n');
    ss_algoVEM.ComputeExpectError;
    
    fprintf('Update Noise\n');
    ss_algoVEM.UpdateNoiseVar;
    
    ss_algoVEM.ComputeVarFreeEnergy(0);
    varFreeEnergy(iter) = ss_algoVEM.varFreeEnergy;

    fprintf('variational free energy: %.2f\n', varFreeEnergy(iter)/(I*T));
    
    
end
ss_algoVEM.EStepPCGComplex;


%% Compute estimated sources

Ls = T_orig-La+1;
se = iofstft_multi(ss_algoVEM.m_post_r + 1i*ss_algoVEM.m_post_i, Ls, win, hop_percent);

for j=1:J
    audiowrite(fullfile(resultsDir, [resultsPrefix num2str(j) '.wav']), se(:,j),fs);
end
