function [ x ] = iofstft_multi( X, siglen, win, hop_percent, b_fft )

%IOFSTFT_MULTI Multichannel Inverse Odd-Frequency Short-Term Fourier Transform (IOFSTFT)
%
% x = iofstft_multi( X, siglen)
% x = iofstft_multi( X, siglen, win)
% x = iofstft_multi( X, siglen, win, hop_percent)
% x = iofstft_multi( X, siglen, win, hop_percent, b_fft )
%
% Inputs:
% X: [F x N x J] matrix containing the OFSTFT coefficients for J
% sources with F frequency bins and N time frames or [F x N x J x I]
% matrix containing the OFSTFT coefficients of J spatial
% source images over I channels.
% siglen: number of samples to which the corresponding time-domain signals
% are to be truncated.
% win: [1 x wlen] synthesis window
% hop_percent: hop size as a percentage of the window length
% b_fft: if 1, fft is used to compute the iOFSTFT.
%
% Output:
% x: [siglen x J] matrix or [siglen x J x I] matrix containing the
% corresponding time-domain signals.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Simon Leglaive, Tlcom ParisTech, 2017
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if nargin<2
    error('Not enough input arguments.');
end
if nargin<3
    wlen = 2*size(X,1); % window length
    win = sin((.5:wlen-.5)/wlen*pi); % sine window
end
if nargin<4
    hop_percent = 0.5;
end
if nargin<5
    b_fft = 0;
end

[F,N,J,I] = size(X);

X = cat(1,X,conj(X(end:-1:1,:,:,:)));

wlen = 2*F; % window length
hop = wlen*hop_percent; 
x = zeros((N-1)*hop + wlen, J, I);

if ~b_fft
    
    T = zeros(wlen,wlen);
    for f = 0:wlen-1,
        T(f+1,:) = win.*( sqrt(1/wlen) * exp(-1i*2*pi*(0:wlen-1)*(f+0.5)/wlen) );
    end
    
    for i=1:I
        for j = 1:J
            frames = real(T'*X(:,:,j,i));
            for n = 1:N
                frameInd = (n-1)*hop + (1:wlen);
                x(frameInd,j,i) = x(frameInd,j,i) + frames(:,n);
            end
        end
    end
else
    for i=1:I
        for j = 1:J
            y = sqrt(wlen)*real(ifft(X(:,:,j,i)).*(exp(1i*pi*(0:wlen-1)/wlen).'*ones(1,N)));
            for n = 1:N
                frameInd = (n-1)*hop + (1:wlen);
                x(frameInd,j,i) = x(frameInd,j,i) + win'.*y(:,n);
            end
        end
    end
end

x = x(wlen*(1-hop_percent)+1:wlen*(1-hop_percent)+siglen,:,:);

end

