function [ X ] = ofstft_multi( x, win, hop_percent, b_fft )

% OFSTFT_MULTI Multichannel Odd-Frequency Short-Term Fourier Transform (OFSTFT)
%
% X=ofstft_multi(x)
% X=ofstft_multi(x,win)
% X=ofstft_multi(x,win, hop_percent)
% X=ofstft_multi(x,win, hop_percent, b_fft)
%
% Inputs:
% x: [siglen x I] matrix containing the I channels time-domain mixture signals
% with siglen samples.
% win: [1 x wlen] analysis window
% hop_percent: hop size as a percentage of the window length
% b_fft: if 1, fft is used to compute OFSTFT.
%
% Output:
% X: [F x N x I] matrix containing the OFSTFT coefficients with F
% frequency bins and N time frames.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Simon Leglaive, Tlcom ParisTech, 2017
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if nargin<1
    error('Not enough input arguments.');
end
if nargin<2
    wlen = 1024; % window length
    win = sin((.5:wlen-.5)/wlen*pi); % sine window
    b_fft = 0;
end
if nargin<3
    hop_percent = 0.5;
end
if nargin<4
    b_fft = 0;
end

if size(x,2) > size(x, 1)
    error('The signals must be within columns.');
end

% window length
wlen = length(win);

% number of frequency bins = wlen/2
F = wlen/2;

% hop size
hop = wlen*hop_percent;

% zero padding for having constant overlap add of the window at the
% begining of the signal to be analysed
x = [zeros(wlen*(1-hop_percent),size(x,2)) ; x];
[T,I] = size(x);

% number of frames
N = fix(T/hop) + (mod(T, hop) ~= 0); % add one frame if T is not multiple of hop

% zero padding for handling last frame
T_zp = (N-1)*hop + wlen;
Nzp_end = T_zp - T;
x = [x ; zeros(Nzp_end, I)];

% compute sample indices for each frame
frameInd = zeros(wlen,N); % column n of frameInd contains the indices of the signal to be considered for frame n
for n=1:N
    frameInd(:,n) = ((n-1)*hop+1 : (n-1)*hop + wlen)';
end

% compute OFSTFT
X = zeros(F,N,I);

if ~b_fft
    
    T = zeros(F,wlen);
    for f = 0:F-1,
        T(f+1,:) = win.*( sqrt(1/wlen) * exp(-1i*2*pi*(0:wlen-1)*(f+0.5)/wlen) );
    end
    
    for i = 1:I
        tmp = x(:,i);
        X(:,:,i) = T*tmp(frameInd);
    end
    
else
    
    for i = 1:I
        x_i = x(:,i);
        frameMat = x_i(frameInd);
        frameMat = frameMat.*repmat(win',1,N);
        framefftMat = fft(frameMat.*(exp(-1i*pi*(0:wlen-1)/wlen).'*ones(1,N)));
        X(:,:,i) = sqrt(1/wlen)*framefftMat(1:F,:);
    end
    
end

end

